
--- Camera module that supports zooming, rotation, bounds and entity following.
-- @todo camera effects

local random, floor = math.random, math.floor
local push, pop, translate, scale = love.graphics.push, love.graphics.pop, love.graphics.translate, love.graphics.scale
local setColor, rectangle = love.graphics.setColor, love.graphics.rectangle


local Camera = {}


--- New world.
function Camera:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self
  o:init()
  return o
end


--- Init the world.
function Camera:init()
  assert(type(self.x) == 'number', 'required arg x must be a number')
  assert(type(self.y) == 'number', 'required arg y must be a number')
  assert(type(self.w) == 'number', 'required arg w must be a number')
  assert(type(self.h) == 'number', 'required arg h must be a number')

  --- New camera data table.
  -- @field x x position
  -- @field y y position
  -- @field xo x offset, not affeted by bounds and follow
  -- @field yo y offset, not affeted by bounds and follow
  -- @field w width, affects bounds and follow
  -- @field h height, affects bounds and follow
  -- @field zoom zoom factor, < 1 further, > 1 closer
  -- @field bounds camera bounds table
  -- @field follow camera follow table
  -- @table camera.camera

  self.xo = 0
  self.yo = 0
  self.zoom = 1

  --- Camera bounds.
  -- Affects bounded sets and follow.
  -- @field x1 top left corner x
  -- @field y1 top left corner y
  -- @field x2 bottom right corner x
  -- @field y2 botom right corner y
  self.bounds = {
    x1 = 0,
    y1 = 0,
    x2 = 0,
    y2 = 0
  }

  --- Follow factor.
  -- 0 to 1, 0 is center, 1 is camera edges.
  -- @field x x follow factor
  -- @field y y follow factor
  self.followFactor = {
    x = 0,
    y = 0
  }

  self.xnoise, self.ynoise = {}, {}
end


-- Clamp a value between min and max.
-- @number val the value
-- @number min the clamp minimum
-- @number max the clamp maximum
function Camera:clamp(val, min, max)
  return val < min and min or (val > max and max or val)
end


--- Gect the world coordinates at camera x, y
-- @tparam camera camera camera data table
-- @tparam number x camera x
-- @tparam number y camera y
-- @treturn number world x
-- @treturn number world y
function Camera:toWorld(x, y)
  return (x * self.zoom) + self.x + self.xo, (y / self.zoom) + self.y + self.yo
end


--- Get the camera coordinates at world x, y
-- @tparam camera camera camera data table
-- @tparam number x world x
-- @tparam number y world y
-- @treturn number camera x
-- @treturn number camera y
function Camera:toCamera(x, y)
  return (x - self.x - self.xo) * self.zoom, (y - self.y - self.yo) * self.zoom
end


--- Set the camera x location and keep it in the camera bounds.
-- Does not affect camera xo.
-- @tparam camera camera camera data table
-- @number x x location
function Camera:setBoundedX(x)
  if (self.bounds.x2 - self.bounds.x1) * self.zoom < self.w then
    self.x = - (self.w / self.zoom - self.bounds.x2 - self.bounds.x1) / 2
  else
    self.x = self:clamp(x, self.bounds.x1, self.bounds.x2 - self.w / self.zoom)
  end
end


--- Set the camera y location and keep it in the camera bounds.
-- Does not affect camera xo.
-- @tparam camera camera camera data table
-- @number y y location
function Camera:setBoundedY(y)
  if (self.bounds.y2 - self.bounds.y1) * self.zoom < self.h then
    self.y = - (self.h / self.zoom - self.bounds.y2 - self.bounds.y1) / 2
  else
    self.y = self:clamp(y, self.bounds.y1, self.bounds.y2 - self.h / self.zoom)
  end
end


--- Follow an entity, or any table with x, y, w, h numeric values.
-- Following is affected by the camera bounds and follow factor.
-- @tparam camera camera camera data table
-- @tparam entity.entity entity the entity to follow
function Camera:follow(entity)
  local fx = self.w / 2 * (1 - self.followFactor.x) / self.zoom
  local fy = self.h / 2 * (1 - self.followFactor.y) / self.zoom
  local ex = entity.x + entity.w / 2
  local ey = entity.y + entity.h / 2

  if ex < (self.x + fx) then
    self:setBoundedX(ex - fx)
  elseif ex > (self.x + self.w / self.zoom - fx) then
    self:setBoundedX(ex - self.w / self.zoom + fx)
  end

  if ey < (self.y + fy) then
    self:setBoundedY(ey - fy)
  elseif ey > (self.y + self.h / self.zoom - fy) then
    self:setBoundedY(ey - self.h / self.zoom + fy)
  end
end


function Camera:shake(duration, frequency, amplitude)
  -- convert milliseonds to seconds
  self.duration = duration / 1000
  self.frequency = frequency
  self.amplitude = amplitude
  self.samples = self.duration * self.frequency
  self.shaking = true
  self.time = 0

  -- 1 extra because we take s and s + 1
  for i = 0, self.samples do
    self.xnoise[i] = random() * 2 - 1
    self.ynoise[i] = random() * 2 - 1
  end
end


-- updates shake effect
function Camera:update(dt)
  if self.shaking then
    self.time = self.time + dt

    if self.time > self.duration then
      self.shaking = false
    else
      local s = self.time * self.frequency
      local s0 = floor(s)
      local s1 = s0 + 1
      local tf = (self.duration - self.time) / self.duration * self.amplitude

      self.xo = (self.xnoise[s0] + (s - s0) * (self.xnoise[s1] - self.xnoise[s0])) * tf
      self.yo = (self.ynoise[s0] + (s - s0) * (self.ynoise[s1] - self.ynoise[s0])) * tf
    end
  end
end


--- Push the camera transformation state on to the transformation stack.
-- The camera location and zoom affect the transformation state.
-- @tparam camera camera camera data table
function Camera:push()
  push()
  scale(self.zoom, self.zoom)
  translate(-(self.x + self.xo), -(self.y + self.yo))
end


function Camera:pop()
  pop()
end


--- Debug draw of the camera follow region and bounds.
-- @tparam camera camera camera data table
function Camera:drawDebug()
  self:push()

  local fx = self.w / 2 * (1 - self.followFactor.x) / self.zoom
  local fy = self.h / 2 * (1 - self.followFactor.y) / self.zoom
    -- follow
  setColor(200, 150, 0)
  rectangle('line', self.x + fx, self.y + fy,
    self.w / self.zoom - fx - fx, self.h / self.zoom - fy - fy)

  --- bounds
  setColor(255, 50, 0)
  rectangle('line', self.bounds.x1 + 1, self.bounds.y1 + 1,
    self.bounds.x2 - self.bounds.x1 - 2, self.bounds.y2 - self.bounds.y1 - 2)

  self:pop()
end


return Camera
