
--- Vector module.
-- Coordinate system orientated the same way as Love2D.
-- 0, 0 on the top left of the screen.
-- X increases to the right, Y increases downwards.
-- Rotation clockwise.


local rad, sqrt, cos, sin, atan2 = math.rad, math.sqrt, math.cos, math.sin, math.atan2

local rad90, rad360 = rad(90), rad(360)


local vector = {}


--- Component wise addition.
-- @tparam number x1
-- @tparam number y1
-- @tparam number x2
-- @tparam number y2
-- @treturn number x
-- @treturn number y
function vector.add(x1, y1, x2, y2)
  return x1 + x2, y1 + y2
end


--- Component wise subtraction.
-- @tparam number x1
-- @tparam number y1
-- @tparam number x2
-- @tparam number y2
-- @treturn number x
-- @treturn number y
function vector.sub(x1, y1, x2, y2)
  return x1 - x2, y1 - y2
end


--- Scalar multiplcation.
-- @tparam number x
-- @tparam number y
-- @tparam number s
-- @treturn number x
-- @treturn number y
function vector.mul(x, y, s)
  return x * s, y * s
end


--- Scalar division.
-- @tparam number x
-- @tparam number y
-- @tparam number s
-- @treturn number x
-- @treturn number y
function vector.div(x, y, s)
  return x / s, y / s
end


--- Dot product.
-- @tparam number x1
-- @tparam number y1
-- @tparam number x2
-- @tparam number y2
-- @treturn number
function vector.dot(x1, y1, x2, y2)
  return x1 * x2 + y1 * y2
end


--- Cross product.
-- @tparam number x1
-- @tparam number y1
-- @tparam number x2
-- @tparam number y2
-- @treturn number
function vector.cross(x1, y1, x2 ,y2)
  return x1 * y2 - y1 * x2
end


--- Length.
-- @tparam number x
-- @tparam number y
-- @treturn number
function vector.len(x, y)
  return sqrt(x * x + y * y)
end


--- Length squared.
-- @tparam number x
-- @tparam number y
-- @treturn number
function vector.len2(x, y)
  return x * x + y * y
end


--- Distance.
-- @tparam number x1
-- @tparam number y1
-- @tparam number x2
-- @tparam number y2
-- @treturn number
function vector.dist(x1, y1, x2, y2)
  return vector.len(x1 - x2, y1 - y2)
end


--- Distance squared.
-- @tparam number x1
-- @tparam number y1
-- @tparam number x2
-- @tparam number y2
-- @treturn number
function vector.dist2(x1, y1, x2, y2)
  return vector.len2(x1 - x2, y1 - y2)
end


--- Unit vector.
-- @tparam number x
-- @tparam number y
-- @treturn number x
-- @treturn number y
function vector.unit(x, y)
  local len = vector.len(x, y)
  if len == 0 then return x, y end
  return x / len, y / len
end


--- Trim vector at max length.
-- @tparam number x
-- @tparam number y
-- @tparam number max
-- @treturn number x
-- @treturn number y
function vector.trim(x, y, max)
  local len = vector.len(x, y)
  if len <= max then return x, y end
  return x / len * max, y / len * max
end


--- Get a vector perpendicular to the left.
-- @tparam number x
-- @tparam number y
-- @treturn number x
-- @treturn number y
function vector.left(x, y)
  return -y, x
end


--- Get a vector perpendicular to the right.
-- @tparam number x
-- @tparam number y
-- @treturn number x
-- @treturn number y
function vector.right(x, y)
  return y, -x
end


--- Project a vector onto another.
-- @tparam number x1
-- @tparam number y1
-- @tparam number x2
-- @tparam number y2
-- @treturn number x
-- @treturn number y
function vector.project(x1, y1, x2, y2)
  local s = (x1 * x2 + y1 * y2) / (x2 * x2 + y2 * y2)
  return s * x2, s * y2
end


--- Mirro a vector onto another.
-- @tparam number x1
-- @tparam number y1
-- @tparam number x2
-- @tparam number y2
-- @treturn number x
-- @treturn number y
function vector.mirror(x1, y1, x2, y2)
  local s = 2 * (x1 * x2 + y1 * y2) / (x2 * x2 + y2 * y2)
  return s * x2 - x1, s * y2 - y1
end


--- Get the angle in radians of a vector.
-- @tparam number x
-- @tparam number y
-- @treturn number radians
function vector.angle(x, y)
  local angle = -(-atan2(y, x) - rad90)
  if angle < 0 then
    return rad360 + angle
  elseif angle == -0 then
    return 0
  else
    return angle
  end
end


--- Rotate a vector by an angle in radians.
-- @tparam number x
-- @tparam number y
-- @tparam number angle
-- @treturn number x
-- @treturn number y
function vector.rotate(x, y, angle)
  local c, s = cos(angle - rad90), sin(angle - rad90)
  return c * x - s * y, s * x + c * y
end


--- Get a vector from an angle in radians with an optional magnitude.
-- @tparam number angle
-- @tparam[opt] number mag
-- @treturn number x
-- @treturn number y
function vector.from(angle, mag)
  return cos(angle - rad90) * (mag or 1), sin(angle - rad90) * (mag or 1)
end


return vector
