
--- World module.
-- SupportileSize loading an empty world or Tiled, image and csv maps.
-- Entity collisions and callbacks with tiles, entities and successful moves.
-- Entity zoning for improved collision performance.
-- Entity pooling for improved entity recycling.
-- @classmod World

local Tile = require('tile')
local Entity = require('entity')
local Fighter = require('entity.fighter')
local Rogue = require('entity.rogue')
local Ice = require('entity.ice')
local Sword = require('entity.sword')
local Armor = require('entity.armor')
local Boots = require('entity.boots')
local Joystick = require('joystick')
local game = require('game')

local floor, ceil, max, min = math.floor, math.ceil, math.max, math.min
local setColor, rectangle, gPrint = love.graphics.setColor, love.graphics.rectangle, love.graphics.print


local World = {}


--- axis enum
World.axis = {
  X = 1,
  Y = 2
}

--- direction enum
World.direction = {
  U = 1,
  D = 2,
  L = 3,
  R = 4
}


function World:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self
  o:init()
  return o
end


-- empty init
function World:init()
end


-- load the world.
function World:load()

  --- camera
  -- @field self.camera

  --- width
  -- @field self.w

  --- height
  -- @field self.h

  --- tile size
  -- @field self.tileSize

  --- zone tiles
  -- @field self.zoneTiles

  --- number of zones across
  self.zoneWidth = ceil(self.w / self.zoneTiles)
  --- number of zones down
  self.zoneHeight = ceil(self.h / self.zoneTiles)
  --- zone size in pixels
  self.zoneSize = self.zoneTiles * self.tileSize
  --- 2 dimensional array of tiles
  self.tiles = {}
  --- entities
  self.entities = {}
  self.players = {}
  --- 2 dimensional array of entities
  self.zones = {}

  -- init zones
  for x = 0, self.zoneWidth - 1 do
    self.zones[x] = {}
    for y = 0, self.zoneHeight - 1 do
      self.zones[x][y] = {}
    end
  end
end




--- New world.
function World:newEmpty(screen, w, h, tileSize, zoneSize)
  -- asserts

  math.randomseed(os.clock())

  local world = self:new({
    screen = screen,
    camera = screen.camera,
    w = w,
    h = h,
    tileSize = tileSize,
    zoneTiles = zoneTiles or 3,
    sheet = game:sheet('image/all')
  })

  world.addCreepTime = 0
  world.addCreepTimeTotal = 0.4

  local tileset = 'texture_ice'

  world:load()

  for x = 0, world.w - 1 do
    world.tiles[x] = {}
    for y = 0, world.h - 1 do
      world.tiles[x][y] = Tile:new({ x = x, y = y})

      world.tiles[x][y].quad = world.sheet.parts[tileset].quads[1][2]

      local isBorder = false

      -- top
      if y == 0 then
        world.tiles[x][y].collidable = true
        world.tiles[x][y].quad = world.sheet.parts[tileset].quads[1][0]
        isBorder = true
      end

      -- top - 1
      if y == 1 then
        world.tiles[x][y].collidable = true
        world.tiles[x][y].quad = world.sheet.parts[tileset].quads[1][1]
        isBorder = true
      end

      -- bottom
      if y == world.h - 1 then
        world.tiles[x][y].collidable = true
        world.tiles[x][y].quad = world.sheet.parts[tileset].quads[1][3]
        isBorder = true
      end

      -- left
      if x == 0 then
        world.tiles[x][y].collidable = true
        world.tiles[x][y].quad = world.sheet.parts[tileset].quads[0][2]
        isBorder = true
      end

      -- right
      if x == world.w - 1 then
        world.tiles[x][y].collidable = true
        world.tiles[x][y].quad = world.sheet.parts[tileset].quads[2][2]
        isBorder = true
      end

      -- top left
      if x == 0 and y == 0 then
        world.tiles[x][y].quad = world.sheet.parts[tileset].quads[0][0]
      end

      -- bottom right
      if x == world.w - 1 and y == world.h - 1 then
        world.tiles[x][y].quad = world.sheet.parts[tileset].quads[2][3]
      end

      -- bottom left
      if x == 0 and y == world.h - 1 then
        world.tiles[x][y].quad = world.sheet.parts[tileset].quads[0][3]
      end

      -- top right
      if x == world.w - 1 and y == 0 then
        world.tiles[x][y].quad = world.sheet.parts[tileset].quads[2][0]
      end

      if not isBorder then
        if math.random() > 0.9 then
          world:addEntity(Ice:new({
            x = x * world.tileSize,
            y = y * world.tileSize,
          }))
        end
      end

    end
  end

  -- -- random gen
  -- world.tiles[4][5].quad = world.sheet.parts[tileset].quads[4][1]
  -- world.tiles[4][5].collidable = true
  -- world.tiles[4][6].quad = world.sheet.parts[tileset].quads[4][2]
  -- world.tiles[4][6].collidable = true

  if love.joystick.getJoystickCount() > 0 then
    screen.player1 = Fighter:new({
      world = world,
      x = 6 * world.tileSize,
      y = 6 * world.tileSize,
      joystick = Joystick:new({ joystick = love.joystick.getJoysticks()[1] })
    })
    world:addEntity(screen.player1)
    world.players[1] = screen.player1
  end

  if love.joystick.getJoystickCount() > 1 then
    screen.player2 = Fighter:new({
      world = world,
      x = 10 * world.tileSize,
      y = 5 * world.tileSize,
      joystick = Joystick:new({ joystick = love.joystick.getJoysticks()[2] })
    })
    world:addEntity(screen.player2)
    world.players[2] = screen.player2
  end

  if love.joystick.getJoystickCount() > 2 then
    screen.player3 = Fighter:new({
      world = world,
      x = 8 * world.tileSize,
      y = 6 * world.tileSize,
      joystick = Joystick:new({ joystick = love.joystick.getJoysticks()[3] })
    })
    world:addEntity(screen.player3)
    world.players[3] = screen.player3
  end

  -- world:addEntity(Sword:new({
  --   x = 9 * world.tileSize + 4,
  --   y = 9 * world.tileSize + 4
  -- }))

  -- world:addEntity(Armor:new({
  --   x = 13 * world.tileSize + 4,
  --   y = 10 * world.tileSize + 4
  -- }))

  -- world:addEntity(Boots:new({
  --   x = 16 * world.tileSize + 4,
  --   y = 6 * world.tileSize + 4
  -- }))


  return world
end


-- Create a world from a lua map file.
-- Call load_csv_tile callback for each tile.
-- @tparam world world world data table
function World:newMap(name, camera)
  -- asserts

  local mapping = require('map')
  local map = require('map.' .. name)
  local sheet = game:sheet('image/all')

  self.quad = love.graphics.newQuad(2, 2, 28, 28, 256, 256)

  local x, y, w, h = sheet.parts.player.quad:getViewport()

  local world = self:new({
    camera = camera,
    w = map.tiles[1]:len(),
    h = #map.tiles,
    tileSize = map.tileSize,
    zoneTiles = map.zoneTiles or 3,
    mapping = mapping,
    map = map,
    sheet = sheet
  })

  world:load()

  -- init tiles
  for x = 0, world.w - 1 do
    world.tiles[x] = {}
    -- y done later
  end

  -- named mapping access
  mapping.name = {}
  for k, tile in pairs(mapping) do
    if tile.name then
      if mapping.name[tile.name] then
        error('tile name "' .. tile.name .. '" exists')
      else
        mapping.name[tile.name] = tile
      end
    end
  end

  -- tiles
  for y = 0, world.h - 1 do
    for x = 0, world.w - 1 do
      local char = map.tiles[y + 1]:sub(x + 1, x + 1)
      local tile = Tile:new({ x = x, y = y})
      if mapping[char] and sheet.parts[mapping[char].tileset] then
        tile.quad = sheet.parts[mapping[char].tileset].quads[mapping[char].x][mapping[char].y]
      end
      world.tiles[x][y] = world:loadTile(tile, char, mapping[char])
    end
  end

  return world
end


--- Load tile callback for world.load function.
-- @tparam world world world data table
-- @tparam tile tile tile data table
-- @treturn tile tile tile data table
function World:loadTile(tile, char, mapping)
  return tile
end


--- Add entity to world.
function World:addEntity(e)
  e.world = self
  self.entities[e] = e
end


--- Test if the entity collides with any entities.
-- If result is true then return the first entity.
-- If result if false then resolve the entity collision with e1.
function World:entityCollision(e1, nx, ny, axis, result)
  local collision = false

  -- determine axis direction
  if axis == self.axis.Y then
    if ny - e1.y < 0 then dir = self.direction.U else dir = self.direction.D end
  else
    if nx - e1.x < 0 then dir = self.direction.L else dir = self.direction.R end
  end

  -- loop through entity zones
  for k, zone in pairs(e1.zones) do
    -- loop through zone entities
    for ek, e2 in pairs(zone) do
      if e2.remove then break end

      if e1 ~= e2 then
        if
        nx < e2.x + e2.w and e2.x < nx + e1.w
        and
        ny < e2.y + e2.h and e2.y < ny + e1.h
        then
          collision = true

          if result then
            return e2
          else
            e1.entityCollision(e1, e2, nx, ny, dir)
          end
        end
      end -- if not the same entity
    end -- entities
  end -- zones

  return collision
end


--- Test if the entity collides with any tiles.
-- If param result is true then return the first tile.
-- If param result if false then resolve the tile collision with the entity.
-- Also performs world boundary checks and collision callbacks.
-- @tparam entity.entity e entity data table
-- @tparam number nx entity new x position
-- @tparam number ny entity new y position
-- @tparam World.axis axis the axis the collision check is being performed on
-- @tparam boolean result return the tile or trigger the callback
-- @treturn[1] tile if result then the tile or nil
-- @treturn[2] boolean if not result then the tile or nil
function World:tileCollision(e, nx, ny, axis, result)
  if e.types.attack == true then
    return false
  end

  local collision = false

  -- determine axis direction
  if axis == self.axis.Y then
    if ny - e.y < 0 then dir = self.direction.U else dir = self.direction.D end
  else
    if nx - e.x < 0 then dir = self.direction.L else dir = self.direction.R end
  end

  -- determine affected tiles
  local x1 = floor(min(e.x, nx) / self.tileSize)
  local y1 = floor(min(e.y, ny) / self.tileSize)
  local x2 = floor((max(e.x, nx) + e.w - 0.1) / self.tileSize)
  local y2 = floor((max(e.y, ny) + e.h - 0.1) / self.tileSize)

  -- boundary checks
  if x1 < 0 or x2 < 0 then
    if result then return nil else
      e.tileCollision(e, nil, -1, y1, nx, ny, dir)
      return true
    end
  elseif x1 > self.w - 1 or x2 > self.w - 1 then
    if result then return nil else
      e.tileCollision(e, nil, self.w, y1, nx, ny, dir)
      return true
    end
  elseif y1 < 0 or y2 < 0 then
    if result then return nil else
      e.tileCollision(e, nil, x1, -1, nx, ny, dir)
      return true
    end
  elseif y1 > self.h - 1 or y2 > self.h - 1 then
    if result then return nil else
      e.tileCollision(e, nil, 0, self.h, nx, ny, dir)
      return true
    end
  end

  -- tile checks
  for x = x1, x2 do
    for y = y1, y2 do
      if self.tiles[x][y].collidable then
        collision = true
        if result then return self.tiles[x][y] else
          e.tileCollision(e, self.tiles[x][y], x, y, nx, ny, dir)
        end
      end
    end
  end

  return collision
end


--- Try to move an entity to nx and ny on one axis.
-- If the X axis is being tested there should be no change in y and vice versa.
-- Entity collisions are tested first and tile collisions after.
-- If no collisions occurred the move will be resolved.
-- @tparam entity.entity e entity data table
-- @tparam number nx entity new x position
-- @tparam number ny entity new y position
-- @tparam world.axis axis the axis the collision check is being performed on
function World:moveEntityAxis(e, nx, ny, a)
  if
    not self:tileCollision(e, nx, ny, a) and
    not self:entityCollision(e, nx, ny, a)
  then
    e:move(nx, ny)
  end

end


--- Try to move an entity on the world to nx and ny.
-- Tries the movement on X axis and then Y axis.
-- @tparam entity.entity e entity data table
-- @tparam number nx entity new x position
-- @tparam number ny entity new y position
function World:moveEntity(e, nx, ny)
  self:moveEntityAxis(e, nx, e.y, self.axis.X)
  self:moveEntityAxis(e, e.x, ny, self.axis.Y)
end


function World:pixelsToTiles(pixels)
  return floor(pixels / self.tileSize)
end


function World:tilesToPixels(tiles)
  return tiles * self.tileSize
end


function World:worldToTile(x, y)
  local x1 = floor(x / self.tileSize)
  local x2 = floor(y / self.tileSize)

  if x1 >= 0 and x1 < self.w - 1 and y1 >= 0 and y1 < self.h - 1 then
    return self.tiles[x1][y1]
  else
    return nil
  end
end


function World:tileAbove(e, distance)
  return tileCollision(e, e.x, e.y - distance, axis.Y, true)
end


function World:tileBelow(e, distance)
  return tileCollision(e, e.x, e.y + distance, axis.Y, true)
end


function World:tileLeft(e, distance)
  return tileCollision(e, e.x - distance, e.y, axis.X, true)
end


function World:tileRight(e, distance)
  return tileCollision(e, e.x + distance, e.y, axis.X, true)
end


function World:entityAbove(e, distance)
  return entityCollision(e, e.x, e.y - distance, axis.Y, true)
end


function World:entityBelow(e, distance)
  return entityCollision(e, e.x, e.y + distance, axis.Y, true)
end


function World:entityLeft(e, distance)
  return entityCollision(e, e.x - distance, e.y, axis.X, true)
end


function World:entityRight(e, distance)
  return entityCollision(e, e.x + distance, e.y, axis.X, true)
end


function World:updateEntityZones(e)
  -- remove
  for key, zone in pairs(e.zones) do
    zone[e] = nil
    e.zones[key] = nil
  end

  -- determine affected zones
  local x1 = floor(e.x / self.zoneSize - 0.1)
  if x1 < 0 then x1 = 0 end
  local y1 = floor(e.y / self.zoneSize - 0.1)
  if y1 < 0 then y1 = 0 end
  local x2 = floor((e.x + e.w - 0.1) / self.zoneSize)
  if x2 > self.zoneWidth - 1 then x2 = self.zoneWidth - 1 end
  local y2 = floor((e.y + e.h - 0.1) / self.zoneSize)
  if y2 > self.zoneHeight - 1 then y2 = self.zoneHeight - 1 end

  -- mark zones
  for zx = x1, x2 do
    for zy = y1, y2 do
      self.zones[zx][zy][e] = e
      e.zones[self.zones[zx][zy]] = self.zones[zx][zy]
    end
  end
end


function World:update(dt)

  self.addCreepTime = self.addCreepTime + dt

  if self.addCreepTime > self.addCreepTimeTotal then
    self:addEntity(Ice:new({
      x = love.math.random(1, self.w - 1) * self.tileSize,
      y = love.math.random(1, self.h - 1) * self.tileSize,
    }))
    self.addCreepTime = 0
  end

  for k, e in pairs(self.entities) do
    if e.remove then
      self.entities[e] = nil

      for key, zone in pairs(e.zones) do
        zone[e] = nil
        e.zones[key] = nil
      end
    else
      e:update(dt)
      self:updateEntityZones(e)
    end
  end
end


function World:drawBackground()

end


function World:drawTiles(x1, y1, x2, y2)
  for x = x1, x2 do
    for y = y1, y2 do
      if self.tiles[x][y].quad then
        self.sheet.sb:add(self.tiles[x][y].quad, x * self.tileSize, y * self.tileSize)
      end
    end
  end
end


function World:drawZones(x1, y1, x2, y2)
  for x = x1, x2 do
    for y = y1, y2 do
      for k, e in pairs(self.zones[x][y]) do
        e.draw(e)
      end
    end
  end
end


--- Default draw. Draws viewport tiles then viewport entity zones.
-- @tparam world world world data table
-- @tparam camera.camera camera camera data table
function World:draw()
  self.camera:push()
  self.sheet.sb:clear()

  -- camera combined position
  local cx = self.camera.x + self.camera.xo
  local cy = self.camera.y + self.camera.yo

  self:drawBackground()

  -- tiles viewport
  self:drawTiles(
    floor(max(cx, 0) / self.tileSize),
    floor(max(cy, 0) / self.tileSize),
    min(floor((cx + self.camera.w / self.camera.zoom) / self.tileSize), self.w - 1),
    min(floor((cy + self.camera.h / self.camera.zoom) / self.tileSize), self.h - 1)
  )

  -- zones viewport
  -- self:drawZones(
  --   floor(max(cx, 0) / self.zoneSize),
  --   floor(max(cy, 0) / self.zoneSize),
  --   min(floor((cx + self.camera.w / self.camera.zoom) / self.zoneSize), self.zoneWidth - 1),
  --   min(floor((cy + self.camera.h / self.camera.zoom) / self.zoneSize), self.zoneHeight - 1)
  -- )

  for k, e in pairs(self.entities) do
    if not e.types.player and not e.remove then
      e:draw()
    end
  end

  love.graphics.draw(self.sheet.sb)

  for i = 1, #self.players do
    local p = self.players[i]
    if p.dashCombo then
      love.graphics.setColor(255, 255, 255, 255 - 255 * p.dashComboTime / p.dashComboTotal)

      if p.dashes.p2 then
        love.graphics.line(p.dashes.p1.x, p.dashes.p1.y, p.dashes.p2.x, p.dashes.p2.y)
      else
        love.graphics.line(p.dashes.p1.x, p.dashes.p1.y, p.x, p.y)
      end
    end
  end

  self.sheet.sb:clear()

  for i = 1, #self.players do
    self.players[i]:draw()
  end

  love.graphics.draw(self.sheet.sb)

  for i = 1, #self.players do
    love.graphics.draw(self.players[i].ps)
  end

  -- love.graphics.setColor(255, 255, 255, 100)
  -- love.graphics.polygon('fill', 100, 100, 200, 100, 150, 200)

  self.camera:pop()
end


function World:drawDebug()
  self.camera:push()

  -- camera combined position
  local cx = self.camera.x + self.camera.xo
  local cy = self.camera.y + self.camera.yo

  -- zones viewport
  local x1 = floor(max(cx, 0) / self.zoneSize)
  local y1 = floor(max(cy, 0) / self.zoneSize)
  local x2 = min(floor((cx + self.camera.w / self.camera.zoom) / self.zoneSize), self.zoneWidth - 1)
  local y2 = min(floor((cy + self.camera.h / self.camera.zoom) / self.zoneSize), self.zoneHeight - 1)

  -- zones
  for x = x1, x2 do
    for y = y1, y2 do
      setColor(200, 100, 100)
      rectangle('line', x * self.zoneSize, y * self.zoneSize, self.zoneSize, self.zoneSize)
      for k, entity in pairs(self.zones[x][y]) do
        setColor(255, 255, 255)
        rectangle('line', entity.x, entity.y, entity.w, entity.h)
      end
    end
  end

  self.camera:pop()

 -- zones
  setColor(255, 255, 255)
  for x = x1, x2 do
    for y = y1, y2 do
      -- top left corner
      local zx = (x * self.zoneSize - self.camera.x - self.camera.xo) * self.camera.zoom
      local zy = (y * self.zoneSize - self.camera.y - self.camera.yo) * self.camera.zoom

      -- zone coordinates
      gPrint(x .. ',' .. y, zx + 5, zy + 5)

      local entityCount = 0

      for k, v in pairs(self.zones[x][y]) do
        entityCount = entityCount + 1
      end

      gPrint(entityCount, zx + 5, zy + 20)
    end
  end

  -- statileSize
  setColor(255, 255, 255)
  gPrint('zones: ' .. (x2 - x1) * (y2 - y1), 5, 20)
end


return World
